@@ -9,10 +9,13 @@ from nbsafety.analysis.mixins import SaveOffAttributesMixin, SkipUnboundArgsMixi
 from nbsafety.data_model.data_symbol import DataSymbol
 from nbsafety.data_model.scope import Scope
 from nbsafety.data_model.timestamp import Timestamp
+from nbsafety.run_mode import ExecutionMode
+from nbsafety.singletons import nbs
+from nbsafety.tracing.mutation_event import resolve_mutating_method
 
 if TYPE_CHECKING:
     from typing import Generator, Iterable, List, Optional, Set, Tuple, Union
-    from nbsafety.types import SymbolRef
+    from nbsafety.types import SupportedIndexType, SymbolRef
 
 logger = logging.getLogger(__name__)
 logger.setLevel(logging.ERROR)
@@ -74,15 +77,18 @@ class ComputeLiveSymbolRefs(SaveOffAttributesMixin, SkipUnboundArgsMixin, VisitL
     # if it is used on the RHS of an assignment
     def visit_Assign_impl(self, targets, value, aug_assign_target=None):
         this_assign_live = set()
-        this_assign_dead = set()
+        # we won't mutate overall dead for visiting simple targets, and we need it to avoid adding false positive lives
         with self.push_attributes(live=this_assign_live):
             self.visit(value)
             if aug_assign_target is not None:
                 self.visit(aug_assign_target)
+        # make a copy, then track the new dead
+        this_assign_dead = set(self.dead)
         with self.push_attributes(dead=this_assign_dead):
             with self.kill_context():
                 for target in targets:
                     self.visit_Assign_target(target)
+        this_assign_dead -= self.dead
         # TODO: ideally under the current abstraction we should
         #  not be resolving static references to symbols here
         if (
@@ -98,11 +104,13 @@ class ComputeLiveSymbolRefs(SaveOffAttributesMixin, SkipUnboundArgsMixin, VisitL
                 for x in (this_assign_dead, (live[0] for live in this_assign_live))
             ]
             if len(lhs) == 1 and len(rhs) == 1:
-                lhs, rhs = [next(iter(x)) for x in (lhs, rhs)]
+                syms: List[DataSymbol] = [next(iter(x)) for x in (lhs, rhs)]
+                lhs_sym, rhs_sym = syms[0], syms[1]
                 # hack to avoid marking `b` as live when objects are same,
                 # or when it was detected that rhs symbol wasn't actually updated
-                if lhs.obj is rhs.obj:
-                    # it's a no-op, so treat it as such
+                if lhs_sym.obj is rhs_sym.obj or lhs_sym.timestamp_excluding_ns_descendents > rhs_sym.timestamp:
+                    # either (a) it's a no-op (so treat it as such), or
+                    #        (b) lhs is newer and it doesn't make sense to refresh
                     this_assign_live.clear()
                     this_assign_dead.clear()
         this_assign_dead -= {live[0] for live in this_assign_live}
@@ -208,6 +216,9 @@ class ComputeLiveSymbolRefs(SaveOffAttributesMixin, SkipUnboundArgsMixin, VisitL
         with self.attrsub_context(inside=False):
             self.visit(node.slice)
 
+    def visit_Delete(self, node: ast.Delete):
+        pass
+
     def visit_GeneratorExp(self, node):
         self.visit_GeneratorExp_or_DictComp_or_ListComp_or_SetComp(node)
 
@@ -252,28 +263,29 @@ def gen_symbols_for_references(
     only_yield_successful_resolutions: bool,
     stmt_counters: Optional[Iterable[int]] = None,
     update_liveness_time_versions: bool = False,
-) -> Generator[Tuple[DataSymbol, bool, bool, Optional[int]], None, None]:
+) -> Generator[Tuple[DataSymbol, Optional[Union[SupportedIndexType, CallPoint]], bool, bool, Optional[int]], None, None]:
     if stmt_counters is None:
         stmt_counters = (None for _ in itertools.count())
     for symbol_ref, stmt_counter in zip(symbol_refs, stmt_counters):
         if isinstance(symbol_ref, str):
             dsym = scope.lookup_data_symbol_by_name(symbol_ref)
             if dsym is not None:
-                yield dsym, False, True, stmt_counter
+                yield dsym, None, False, True, stmt_counter
         elif isinstance(symbol_ref, AttrSubSymbolChain):
             if update_liveness_time_versions:
                 # TODO: only use this branch one staleness checker can be smarter about liveness timestamps.
                 #  Right now, yielding the intermediate elts of the chain will yield false positives in the
                 #  event of namespace stale children.
-                for dsym, is_called, success in scope.gen_data_symbols_for_attrsub_chain(symbol_ref):
+                for dsym, next_ref, is_called, success in scope.gen_data_symbols_for_attrsub_chain(symbol_ref):
                     if only_yield_successful_resolutions and not success:
                         continue
-                    yield dsym, is_called, success, stmt_counter
+                    yield dsym, next_ref, is_called, success, stmt_counter
             else:
-                dsym, is_called, success = scope.get_most_specific_data_symbol_for_attrsub_chain(symbol_ref)
-                if dsym is not None:
+                dsym_et_al = scope.get_most_specific_data_symbol_for_attrsub_chain(symbol_ref)
+                if dsym_et_al is not None:
+                    dsym, next_ref, is_called, success = dsym_et_al
                     if success or not only_yield_successful_resolutions:
-                        yield dsym, is_called, success, stmt_counter
+                        yield dsym, next_ref, is_called, success, stmt_counter
         else:
             logger.warning('invalid type for ref %s', symbol_ref)
             continue
@@ -296,17 +308,54 @@ def get_symbols_for_references(
     return dsyms, called_dsyms
 
 
+def _live_dsym_is_mutating(dsym: DataSymbol, next_ref: CallPoint) -> bool:
+    return resolve_mutating_method(dsym.obj, next_ref.symbol) is not None
+
+
+def _live_dsym_unsafe(dsym: DataSymbol, next_ref: SupportedIndexType) -> bool:
+    if isinstance(dsym.obj, (list, tuple)) and isinstance(next_ref, int) and next_ref >= len(dsym.obj):
+        return True
+    if isinstance(dsym.obj, dict) and next_ref not in dsym.obj:
+        return True
+    if not isinstance(dsym.obj, (dict, list, tuple)) and isinstance(next_ref, str) and not hasattr(dsym.obj, next_ref):
+        # TODO: fix this once we can distinguish between attrs and subscripts in the chain
+        return True
+    return False
+
+
+def _handle_live_symbol(
+    dsym: DataSymbol,
+    next_ref: Optional[Union[SupportedIndexType, CallPoint]],
+    deep_live: Set[DataSymbol],
+    shallow_live: Set[DataSymbol]
+):
+    if next_ref is None:
+        deep_live.add(dsym)
+    elif isinstance(next_ref, CallPoint) and _live_dsym_is_mutating(dsym, next_ref):
+        if nbs().mut_settings.exec_mode == ExecutionMode.NORMAL:
+            shallow_live.add(dsym)
+        elif nbs().mut_settings.exec_mode == ExecutionMode.REACTIVE:
+            return
+        else:
+            raise ValueError('not sure how to handle execution mode %s' % nbs().mut_settings.exec_mode)
+    elif not isinstance(next_ref, CallPoint) and _live_dsym_unsafe(dsym, next_ref):
+        return
+    else:
+        deep_live.add(dsym)
+
+
 def get_live_symbols_and_cells_for_references(
     symbol_refs: Set[Tuple[SymbolRef, int]],
     scope: Scope,
     cell_ctr: int,
     update_liveness_time_versions: bool = False,
-) -> Tuple[Set[DataSymbol], Set[int]]:
-    dsyms: Set[DataSymbol] = set()
+) -> Tuple[Set[DataSymbol], Set[DataSymbol], Set[int]]:
+    deep_dsyms: Set[DataSymbol] = set()
+    shallow_dsyms: Set[DataSymbol] = set()
     called_dsyms: Set[Tuple[DataSymbol, int]] = set()
     only_symbol_refs = (ref[0] for ref in symbol_refs)
     only_stmt_counters = (ref[1] for ref in symbol_refs)
-    for dsym, is_called, success, stmt_ctr in gen_symbols_for_references(
+    for dsym, next_ref, is_called, success, stmt_ctr in gen_symbols_for_references(
         only_symbol_refs,
         scope,
         only_yield_successful_resolutions=False,
@@ -319,20 +368,22 @@ def get_live_symbols_and_cells_for_references(
         if is_called:
             called_dsyms.add((dsym, stmt_ctr))
         else:
-            dsyms.add(dsym)
-    live_from_calls, live_cells = _compute_call_chain_live_symbols_and_cells(
+            _handle_live_symbol(dsym, next_ref, deep_dsyms, shallow_dsyms)
+    deep_live_from_calls, shallow_live_from_calls, live_cells = _compute_call_chain_live_symbols_and_cells(
         called_dsyms, cell_ctr, update_liveness_time_versions
     )
-    dsyms |= live_from_calls
-    return dsyms, live_cells
+    deep_dsyms |= deep_live_from_calls
+    shallow_dsyms |= shallow_live_from_calls
+    return deep_dsyms, shallow_dsyms, live_cells
 
 
 def _compute_call_chain_live_symbols_and_cells(
     live_with_stmt_ctr: Set[Tuple[DataSymbol, int]], cell_ctr: int, update_liveness_time_versions: bool
-) -> Tuple[Set[DataSymbol], Set[int]]:
+) -> Tuple[Set[DataSymbol], Set[DataSymbol], Set[int]]:
     seen = set()
     worklist = list(live_with_stmt_ctr)
-    live = {dsym_stmt[0] for dsym_stmt in live_with_stmt_ctr}
+    deep_live = {dsym_stmt[0] for dsym_stmt in live_with_stmt_ctr}
+    shallow_live: Set[DataSymbol] = set()
     while len(worklist) > 0:
         workitem = worklist.pop()
         if workitem in seen:
@@ -346,17 +397,17 @@ def _compute_call_chain_live_symbols_and_cells(
             cast(ast.FunctionDef, called_dsym.stmt_node).body, init_killed=set(called_dsym.get_definition_args())
         )
         used_time = Timestamp(cell_ctr, stmt_ctr)
-        for dsym, is_called, success, *_ in gen_symbols_for_references(
+        for dsym, next_ref, is_called, success, *_ in gen_symbols_for_references(
             (ref[0] for ref in live_refs), called_dsym.call_scope, only_yield_successful_resolutions=False
         ):
             if is_called:
                 worklist.append((dsym, stmt_ctr))
             if dsym.is_globally_accessible:
-                live.add(dsym)
+                _handle_live_symbol(dsym, next_ref, deep_live, shallow_live)
                 if update_liveness_time_versions:
                     ts_to_use = dsym.timestamp if success else dsym.timestamp_excluding_ns_descendents
                     dsym.timestamp_by_liveness_time_by_cell_counter[cell_ctr][used_time] = ts_to_use
-    return live, {called_dsym.timestamp.cell_num for called_dsym, _ in seen}
+    return deep_live, shallow_live, {called_dsym.timestamp.cell_num for called_dsym, _ in seen}
 
 
 def compute_live_dead_symbol_refs(
