@@ -6,6 +6,11 @@ import logging
 import sys
 from typing import cast, TYPE_CHECKING
 
+try:
+    import numpy
+except ImportError:
+    numpy = None
+
 from nbsafety.data_model.annotation_utils import get_type_annotation, make_annotation_string
 from nbsafety.data_model import sizing
 from nbsafety.data_model.timestamp import Timestamp
@@ -17,7 +22,8 @@ if TYPE_CHECKING:
     from typing import Any, Dict, List, Optional, Set
 
     # avoid circular imports
-    from nbsafety.data_model.scope import Scope, NamespaceScope
+    from nbsafety.data_model.scope import Scope
+    from nbsafety.data_model.namespace import Namespace
 
 logger = logging.getLogger(__name__)
 logger.setLevel(logging.ERROR)
@@ -208,7 +214,7 @@ class DataSymbol:
         if not self.containing_scope.is_namespace_scope:
             return self
         else:
-            containing_scope = cast('NamespaceScope', self.containing_scope)
+            containing_scope = cast('Namespace', self.containing_scope)
             for alias in nbs().aliases[containing_scope.obj_id]:
                 if alias.is_globally_accessible:
                     return alias.get_top_level()
@@ -262,9 +268,9 @@ class DataSymbol:
         return nbs().namespaces.get(self.obj_id, None)
 
     @property
-    def containing_namespace(self) -> Optional[NamespaceScope]:
+    def containing_namespace(self) -> Optional[Namespace]:
         if self.containing_scope.is_namespace_scope:
-            return cast('NamespaceScope', self.containing_scope)
+            return cast('Namespace', self.containing_scope)
         else:
             return None
 
@@ -278,7 +284,16 @@ class DataSymbol:
 
     @property
     def is_garbage(self):
-        return self._tombstone or self.get_ref_count() == 0
+        return self._tombstone
+
+    def is_new_garbage(self):
+        if self._tombstone:
+            return False
+        if numpy is not None and self.containing_namespace is not None and isinstance(self.containing_namespace.obj, numpy.ndarray):
+            # numpy atoms are not interned (so assigning array elts to a variable does not bump refcount);
+            # also seems that refcount is always 0, so just check if the containing namespace is garbage
+            return self.containing_namespace.is_garbage
+        return self.get_ref_count() == 0
 
     @property
     def is_globally_accessible(self):
@@ -286,7 +301,14 @@ class DataSymbol:
 
     @property
     def is_user_accessible(self):
-        return self.is_globally_accessible and not self.is_anonymous
+        return (
+            self.is_globally_accessible
+            and not self.is_anonymous
+            and not (
+                self.containing_namespace is not None
+                and self.containing_namespace.is_anonymous
+            )
+        )
 
     def collect_self_garbage(self):
         """
@@ -314,26 +336,23 @@ class DataSymbol:
 
     def update_obj_ref(self, obj, refresh_cached=True):
         logger.info("%s update obj ref to %s", self, obj)
+        self._tombstone = False
         self._cached_out_of_sync = True
         if nbs().settings.mark_typecheck_failures_unsafe and self.cached_obj_type != type(obj):
             nbs().cell_counters_needing_typecheck |= nbs().cell_counter_by_live_symbol.get(self, set())
-        self._tombstone = False
         self.obj = obj
         if self.cached_obj_id is not None and self.cached_obj_id != self.obj_id:
-            if self.obj_id not in nbs().namespaces:
-                # don't overwrite existing namespace for this obj
-                old_ns = nbs().namespaces.get(self.cached_obj_id, None)
-                if old_ns is not None:
+            new_ns = nbs().namespaces.get(self.obj_id, None)
+            # don't overwrite existing namespace for this obj
+            old_ns = nbs().namespaces.get(self.cached_obj_id, None)
+            if old_ns is not None and (new_ns is None or not new_ns.max_descendent_timestamp.is_initialized) and old_ns.full_namespace_path == self.full_namespace_path:
+                if new_ns is None:
                     logger.info("create fresh copy of namespace %s", old_ns)
                     new_ns = old_ns.fresh_copy(obj)
-                    for dsym in list(old_ns.all_data_symbols_this_indentation(exclude_class=True, is_subscript=False)):
-                        if hasattr(obj, dsym.name):
-                            dsym.update_obj_ref(getattr(obj, dsym.name))
-                            logger.info("shuffle %s from %s to %s", dsym, old_ns, new_ns)
-                            old_ns._data_symbol_by_name.pop(dsym.name, None)
-                            new_ns._data_symbol_by_name[dsym.name] = dsym
-                            dsym.containing_scope = new_ns
-                    # TODO: how best to handle subscripts?
+                else:
+                    new_ns.scope_name = old_ns.scope_name
+                    new_ns.parent_scope = old_ns.parent_scope
+                old_ns.transfer_symbols_to(new_ns)
             self._handle_aliases()
         if refresh_cached:
             self._refresh_cached_obj()
@@ -344,9 +363,14 @@ class DataSymbol:
         self.cached_obj_type = None
 
     def get_ref_count(self):
-        if self.obj is None:
+        if self.obj is None or self.obj is DataSymbol.NULL:
             return -1
-        return sys.getrefcount(self.obj) - 1 - len(nbs().aliases[self.obj_id]) - (self.obj_id in nbs().namespaces)
+        total = sys.getrefcount(self.obj) - 1
+        total -= len(nbs().aliases[self.obj_id])
+        ns = nbs().namespaces.get(self.obj_id, None)
+        if ns is not None and ns.obj is not None and ns.obj is not DataSymbol.NULL:
+            total -= 1
+        return total
 
     def prev_obj_definitely_equal_to_current_obj(self, prev_obj: Optional[Any]) -> bool:
         if prev_obj is None:
@@ -383,6 +407,7 @@ class DataSymbol:
             #  symbol for a given statement. We need a more precise way to determine
             #  the symbol being called than by looking at the stmt in question.
             nbs().statement_to_func_cell[id(stmt_node)] = self
+            self.call_scope = self.containing_scope.make_child_scope(self.name)
         return stmt_node
 
     def _refresh_cached_obj(self):
@@ -478,6 +503,15 @@ class DataSymbol:
             return False
         return True
 
+    def _is_simple_assign(self, new_deps: Set[DataSymbol]) -> bool:
+        if not isinstance(self.stmt_node, (ast.Assign, ast.AnnAssign)):
+            return False
+        if len(new_deps) != 1:
+            return False
+        only_dep: DataSymbol = next(iter(new_deps))
+        # obj ids can get reused for anon symbols like literals
+        return not only_dep.is_anonymous and self.cached_obj_id == only_dep.obj_id
+
     def update_deps(
         self,
         new_deps: Set[DataSymbol],
@@ -489,7 +523,7 @@ class DataSymbol:
         propagate: bool = True,
         refresh: bool = True,
     ) -> None:
-        if self.is_import:
+        if self.is_import and self.obj_id == self.cached_obj_id:
             # skip updates for imported symbols
             # just bump the version if it's newly created
             if mutated or not self._timestamp.is_initialized:
@@ -522,14 +556,15 @@ class DataSymbol:
         equal_to_old = not mutated and self.prev_obj_definitely_equal_to_current_obj(prev_obj)
         if refresh:
             self.refresh(
-                bump_version=not equal_to_old,
+                bump_version=not equal_to_old or type(self.obj) not in DataSymbol.IMMUTABLE_TYPES,
                 # rationale: if this is a mutation for which we have more precise information,
-                # then we don't need to update the ns descendents as this will already have happened
-                refresh_descendent_namespaces=not (mutated and not propagate_to_namespace_descendents),
+                # then we don't need to update the ns descendents as this will already have happened.
+                # also don't update ns descendents for things like `a = b`
+                refresh_descendent_namespaces=not (mutated and not propagate_to_namespace_descendents) and not self._is_simple_assign(new_deps),
                 refresh_namespace_stale=not mutated,
             )
         if propagate and (deleted or not equal_to_old):
-            UpdateProtocol(self)(new_deps, mutated, propagate_to_namespace_descendents)
+            UpdateProtocol(self)(new_deps, mutated, propagate_to_namespace_descendents, refresh)
         self._refresh_cached_obj()
         nbs().updated_symbols.add(self)
         if self.is_class:
