@@ -31,12 +31,14 @@ from nbsafety.ipython_utils import (
 )
 from nbsafety import line_magics
 from nbsafety.data_model.data_symbol import DataSymbol
-from nbsafety.data_model.scope import Scope, NamespaceScope
+from nbsafety.data_model.namespace import Namespace
+from nbsafety.data_model.scope import Scope
 from nbsafety.data_model.timestamp import Timestamp
-from nbsafety.run_mode import SafetyRunMode
+from nbsafety.run_mode import ExecutionMode, SafetyRunMode
 from nbsafety import singletons
 from nbsafety.tracing.safety_ast_rewriter import SafetyAstRewriter
 from nbsafety.tracing.trace_manager import TraceManager
+from nbsafety.utils.misc_utils import GetterPipeline
 
 if TYPE_CHECKING:
     from typing import Any, Dict, Iterable, List, Set, Optional, Tuple, TypeVar, Union
@@ -70,10 +72,13 @@ class MutableNotebookSafetySettings:
     highlights_enabled: bool
     static_slicing_enabled: bool
     dynamic_slicing_enabled: bool
+    exec_mode: ExecutionMode
 
 
 class CheckerResult(NamedTuple):
     live: Set[DataSymbol]
+    deep_live: Set[DataSymbol]
+    shallow_live: Set[DataSymbol]
     live_cells: Set[int]
     live_cells_from_calls: Set[int]
     dead: Set[DataSymbol]
@@ -101,15 +106,17 @@ class NotebookSafety(singletons.NotebookSafety):
             highlights_enabled=kwargs.pop('highlights_enabled', True),
             static_slicing_enabled=kwargs.pop('static_slicing_enabled', True),
             dynamic_slicing_enabled=kwargs.pop('dynamic_slicing_enabled', True),
+            exec_mode=ExecutionMode(kwargs.pop('exec_mode', ExecutionMode.NORMAL)),
         )
         # Note: explicitly adding the types helps PyCharm intellisense
         self.settrace = settrace or sys.settrace
-        self.namespaces: Dict[int, NamespaceScope] = {}
+        self.namespaces: Dict[int, Namespace] = {}
         self.aliases: Dict[int, Set[DataSymbol]] = defaultdict(set)
         self.global_scope: Scope = Scope()
         self.updated_symbols: Set[DataSymbol] = set()
         self.statement_cache: Dict[int, Dict[int, ast.stmt]] = defaultdict(dict)
         self.ast_node_by_id: Dict[int, ast.AST] = {}
+        self.loop_iter_flag_names: Set[str] = set()
         self.cell_id_by_ast_id: Dict[int, CellId] = {}
         self.parent_node_by_id: Dict[int, ast.AST] = {}
         # TODO: we have a lot of fields concerning cells; they should probably get their own
@@ -123,6 +130,9 @@ class NotebookSafety(singletons.NotebookSafety):
         self._typecheck_error_cells: Set[CellId] = set()
         self._counter_by_cell_id: Dict[CellId, int] = {}
         self._cell_id_by_counter: Dict[int, CellId] = {}
+        self.cell_content_by_cell_id = cast(
+            'Dict[CellId, str]', GetterPipeline([self._counter_by_cell_id, self.cell_content_by_counter])
+        )
         self._active_cell_id: Optional[str] = None
         self._run_cells: Set[CellId] = set()
         self.active_cell_position_idx = -1
@@ -142,6 +152,10 @@ class NotebookSafety(singletons.NotebookSafety):
         if use_comm:
             get_ipython().kernel.comm_manager.register_target(__package__, self._comm_target)
 
+    @property
+    def current_cell_content(self):
+        return self.cell_content_by_counter[self.cell_counter()]
+
     @property
     def is_develop(self) -> bool:
         return self.settings.mode == SafetyRunMode.DEVELOP
@@ -235,25 +249,26 @@ class NotebookSafety(singletons.NotebookSafety):
             response = self.check_and_link_multiple_cells(cells_by_id, order_index_by_id)
             response['type'] = 'cell_freshness'
             response['last_cell_exec_position_idx'] = last_cell_exec_position_idx
+            response['exec_mode'] = self.mut_settings.exec_mode.value
+            response['last_executed_cell_id'] = cell_id
+            response['highlights_enabled'] = self.mut_settings.highlights_enabled
             if comm is not None:
                 comm.send(response)
         else:
-            logger.error('Unsupported request type for request %s' % request)
+            dbg_msg = 'Unsupported request type for request %s' % request
+            logger.error(dbg_msg)
+            self._saved_debug_message = dbg_msg
 
     def check_and_link_multiple_cells(
         self,
-        content_by_cell_id: Dict[CellId, str],
+        content_by_cell_id: Optional[Dict[CellId, str]] = None,
         order_index_by_cell_id: Optional[Dict[CellId, int]] = None
     ) -> Dict[str, Any]:
-        if not self.mut_settings.highlights_enabled:
-            return {
-                'stale_cells': [],
-                'fresh_cells': [],
-                'stale_links': {},
-                'refresher_links': {},
-            }
+        if content_by_cell_id is None:
+            content_by_cell_id = self.cell_content_by_cell_id
         stale_cells = set()
         fresh_cells = []
+        new_fresh_cells = []
         stale_symbols_by_cell_id: Dict[CellId, Set[DataSymbol]] = {}
         killing_cell_ids_for_symbol: Dict[DataSymbol, Set[CellId]] = defaultdict(set)
         cell_ids_needing_typecheck = self.get_cell_ids_needing_typecheck()
@@ -292,11 +307,14 @@ class NotebookSafety(singletons.NotebookSafety):
                     # TODO: implement phantom cell usage detection here
                     pass
 
-                if (
-                    self._get_max_timestamp_cell_num_for_symbols(checker_result.live) >
-                    self._counter_by_cell_id.get(cell_id, cast(int, float('inf')))
-                ) and cell_id not in stale_cells and cell_id not in self._typecheck_error_cells:
-                    fresh_cells.append(cell_id)
+                if cell_id not in stale_cells and cell_id not in self._typecheck_error_cells:
+                    max_timestamp_cell_num = self._get_max_timestamp_cell_num_for_symbols(
+                        checker_result.deep_live, checker_result.shallow_live
+                    )
+                    if max_timestamp_cell_num > self._counter_by_cell_id.get(cell_id, cast(int, float('inf'))):
+                        fresh_cells.append(cell_id)
+                    if max_timestamp_cell_num >= self._last_execution_counter:
+                        new_fresh_cells.append(cell_id)
             except SyntaxError:
                 continue
         stale_links: Dict[CellId, Set[CellId]] = defaultdict(set)
@@ -338,6 +356,7 @@ class NotebookSafety(singletons.NotebookSafety):
             #  or at least change the name to a more general "unsafe_cells" or equivalent
             'stale_cells': list(stale_cells | self._typecheck_error_cells),
             'fresh_cells': fresh_cells,
+            'new_fresh_cells': new_fresh_cells,
             'stale_links': {
                 stale_cell_id: list(refresher_cell_ids)
                 for stale_cell_id, refresher_cell_ids in stale_links.items()
@@ -347,19 +366,26 @@ class NotebookSafety(singletons.NotebookSafety):
 
     @staticmethod
     def _get_cell_ast(cell):
+        return ast.parse(NotebookSafety.sanitize_code(cell))
+
+    @staticmethod
+    def sanitize_code(code):
         lines = []
-        for line in cell.strip().split('\n'):
+        for line in code.strip().split('\n'):
             # TODO: figure out more robust strategy for filtering / transforming lines for the ast parser
             # we filter line magics, but for %time, we would ideally like to trace the statement being timed
             # TODO: how to do this?
-            if _NB_MAGIC_PATTERN.search(line) is None:
+            if _NB_MAGIC_PATTERN.search(line.strip()) is None:
                 lines.append(line)
-        return ast.parse('\n'.join(lines))
+        return '\n'.join(lines)
 
-    def _get_max_timestamp_cell_num_for_symbols(self, symbols: Set[DataSymbol]) -> int:
+    @staticmethod
+    def _get_max_timestamp_cell_num_for_symbols(deep_symbols: Set[DataSymbol], shallow_symbols: Set[DataSymbol]) -> int:
         max_timestamp_cell_num = -1
-        for dsym in symbols:
+        for dsym in deep_symbols:
             max_timestamp_cell_num = max(max_timestamp_cell_num, dsym.timestamp.cell_num)
+        for dsym in shallow_symbols:
+            max_timestamp_cell_num = max(max_timestamp_cell_num, dsym.timestamp_excluding_ns_descendents.cell_num)
         return max_timestamp_cell_num
 
     def _build_typecheck_slice(
@@ -390,17 +416,20 @@ class NotebookSafety(singletons.NotebookSafety):
             get_live_symbols_and_cells_for_references(
                 live_symbol_refs, self.global_scope, cell_ctr, update_liveness_time_versions=True
             )
-        live_symbols, live_cells_from_calls = get_live_symbols_and_cells_for_references(
+        deep_live_symbols, shallow_live_symbols, live_cells_from_calls = get_live_symbols_and_cells_for_references(
             live_symbol_refs, self.global_scope, cell_ctr, update_liveness_time_versions=False
         )
+        live_symbols = deep_live_symbols | shallow_live_symbols
         # only mark dead attrsubs as killed if we can traverse the entire chain
         dead_symbols, _ = get_symbols_for_references(
             dead_symbol_refs, self.global_scope, only_yield_successful_resolutions=True
         )
-        stale_symbols = set(dsym for dsym in live_symbols if dsym.is_stale)
+        stale_symbols = {dsym for dsym in live_symbols if dsym.is_stale}
         self.live_symbols_by_cell_counter[cell_ctr] = live_symbols
         return CheckerResult(
             live=live_symbols,
+            deep_live=deep_live_symbols,
+            shallow_live=shallow_live_symbols,
             live_cells={sym.timestamp.cell_num for sym in live_symbols},
             live_cells_from_calls=live_cells_from_calls,
             dead=dead_symbols,
@@ -752,8 +781,6 @@ class NotebookSafety(singletons.NotebookSafety):
             if cell_id is not None:
                 self._counter_by_cell_id[cell_id] = self._last_execution_counter
                 self._cell_id_by_counter[self._last_execution_counter] = cell_id
-
-            if cell_id is not None:
                 # add to run_cells even if it fails precheck, since an attempt
                 # to run indicates that the user is interested in this cell, so
                 # we should add this cell to the set of cells that get checked
@@ -848,6 +875,8 @@ class NotebookSafety(singletons.NotebookSafety):
                 return line_magics.turn_off_warnings_for(line)
             elif cmd == 'turn_on_warnings_for':
                 return line_magics.turn_on_warnings_for(line)
+            elif cmd in ('mode', 'exec_mode'):
+                return line_magics.set_exec_mode(line)
             elif cmd in line_magic_names:
                 logger.warning('We have a magic for %s, but have not yet registered it', cmd)
                 return None
@@ -906,9 +935,8 @@ class NotebookSafety(singletons.NotebookSafety):
 
     def _gc(self):
         # Need to do the garbage check and the collection separately
-        garbage_syms = [dsym for dsym in self.all_data_symbols() if dsym.is_garbage]
+        garbage_syms = [dsym for dsym in self.all_data_symbols() if dsym.is_new_garbage()]
         for dsym in garbage_syms:
-            logger.info('collect sym %s', dsym)
             dsym.collect_self_garbage()
 
     def retrieve_namespace_attr_or_sub(self, obj: Any, attr_or_sub: SupportedIndexType, is_subscript: bool):
