@@ -5,6 +5,7 @@ from collections import defaultdict
 from contextlib import contextmanager
 import functools
 import logging
+import symtable
 import sys
 from typing import cast, TYPE_CHECKING
 
@@ -15,13 +16,23 @@ from nbsafety import singletons
 from nbsafety.analysis.attr_symbols import CallPoint
 from nbsafety.analysis.live_refs import compute_live_dead_symbol_refs
 from nbsafety.data_model.data_symbol import DataSymbol
-from nbsafety.data_model.scope import Scope, NamespaceScope
+from nbsafety.data_model.namespace import Namespace
+from nbsafety.data_model.scope import Scope
 from nbsafety.data_model.timestamp import Timestamp
+from nbsafety.extra_builtins import EMIT_EVENT, TRACING_ENABLED, make_loop_iter_flag_name
 from nbsafety.run_mode import SafetyRunMode
 from nbsafety.singletons import nbs
-from nbsafety.tracing.mutation_event import ArgMutate, ListAppend, ListExtend, ListInsert, StandardMutation
+from nbsafety.tracing.mutation_event import (
+    ArgMutate,
+    ListInsert,
+    ListPop,
+    ListRemove,
+    MutatingMethodEventNotYetImplemented,
+    StandardMutation,
+    resolve_mutating_method,
+)
 from nbsafety.tracing.symbol_resolver import resolve_rval_symbols
-from nbsafety.tracing.trace_events import TraceEvent, EMIT_EVENT
+from nbsafety.tracing.trace_events import TraceEvent
 from nbsafety.tracing.trace_stack import TraceStack
 from nbsafety.tracing.trace_stmt import TraceStatement
 from nbsafety.tracing.utils import match_container_obj_or_namespace_with_literal_nodes
@@ -36,9 +47,9 @@ if TYPE_CHECKING:
     ObjId = int
     MutationCandidate = Tuple[Tuple[Any, Optional[str], Optional[str]], MutationEvent, List[Set[DataSymbol]], List[Any]]
     Mutation = Tuple[int, MutationEvent, Set[DataSymbol], List[Any]]
-    SavedStoreData = Tuple[NamespaceScope, Any, AttrSubVal, bool]
-    SavedDelData = Tuple[NamespaceScope, Any, AttrSubVal, bool]
-    SavedComplexSymbolLoadData = Tuple[NamespaceScope, Any, AttrSubVal, bool, Optional[str]]
+    SavedStoreData = Tuple[Namespace, Any, AttrSubVal, bool]
+    SavedDelData = Tuple[Namespace, Any, AttrSubVal, bool]
+    SavedComplexSymbolLoadData = Tuple[Namespace, Any, AttrSubVal, bool, Optional[str]]
 
 
 logger = logging.getLogger(__name__)
@@ -158,6 +169,7 @@ class BaseTraceManager(singletons.TraceManager):
         else:
             self.sys_tracer = self._make_composed_tracer(self.existing_tracer)
         nbs().settrace(self.sys_tracer)
+        setattr(builtins, TRACING_ENABLED, True)
 
     def _disable_tracing(self, check_enabled=True):
         if check_enabled:
@@ -165,6 +177,7 @@ class BaseTraceManager(singletons.TraceManager):
             assert sys.gettrace() is self.sys_tracer
         self.tracing_enabled = False
         nbs().settrace(self.existing_tracer)
+        setattr(builtins, TRACING_ENABLED, False)
 
     @contextmanager
     def _patch_sys_settrace(self):
@@ -183,8 +196,9 @@ class BaseTraceManager(singletons.TraceManager):
                 self._enable_tracing()
                 yield
         finally:
-            delattr(builtins, EMIT_EVENT)
             self._disable_tracing(check_enabled=False)
+            delattr(builtins, EMIT_EVENT)
+            delattr(builtins, TRACING_ENABLED)
 
     def _should_attempt_to_reenable_tracing(self, frame: FrameType) -> bool:
         return NotImplemented
@@ -255,9 +269,13 @@ class TraceManager(SliceTraceManager):
         self.traced_statements: Dict[NodeId, TraceStatement] = {}
         self.node_id_to_loaded_symbols: Dict[NodeId, List[DataSymbol]] = defaultdict(list)
         self.node_id_to_saved_store_data: Dict[NodeId, SavedStoreData] = {}
+        self.node_id_to_saved_live_subscript_refs: Dict[NodeId, Set[DataSymbol]] = {}
         self.node_id_to_saved_del_data: Dict[NodeId, SavedDelData] = {}
-        self.node_id_to_loaded_literal_scope: Dict[NodeId, NamespaceScope] = {}
+        self.node_id_to_loaded_literal_scope: Dict[NodeId, Namespace] = {}
         self.node_id_to_saved_dict_key: Dict[NodeId, Any] = {}
+        self.cur_cell_symtab: symtable.SymbolTable = symtable.symtable(
+            nbs().sanitize_code(nbs().current_cell_content), f'<cell-{nbs().cell_counter()}>', 'exec'
+        )
 
         self.call_stack: TraceStack = self._make_stack()
         with self.call_stack.register_stack_state():
@@ -269,7 +287,7 @@ class TraceManager(SliceTraceManager):
             # this one gets set regardless of whether tracing enabled
             self.next_stmt_node_id: Optional[NodeId] = None
 
-            self.pending_class_namespaces: List[NamespaceScope] = []
+            self.pending_class_namespaces: List[Namespace] = []
 
             with self.call_stack.needing_manual_initialization():
                 self.cur_frame_original_scope: Scope = nbs().global_scope
@@ -288,7 +306,7 @@ class TraceManager(SliceTraceManager):
             self.lexical_literal_stack: TraceStack = self._make_stack()
             with self.lexical_literal_stack.register_stack_state():
                 # `None` means use 'cur_frame_original_scope'
-                self.active_literal_scope: Optional[NamespaceScope] = None
+                self.active_literal_scope: Optional[Namespace] = None
 
     def module_stmt_counter(self) -> int:
         return self._module_stmt_counter
@@ -352,7 +370,7 @@ class TraceManager(SliceTraceManager):
                 # exception events are followed by return events until we hit an except clause
                 # no need to track dependencies in this case
                 if isinstance(return_to_stmt.stmt_node, ast.ClassDef):
-                    return_to_stmt.class_scope = cast(NamespaceScope, self.cur_frame_original_scope)
+                    return_to_stmt.class_scope = cast(Namespace, self.cur_frame_original_scope)
                 elif isinstance(trace_stmt.stmt_node, ast.Return) or inside_anonymous_call:
                     if not trace_stmt.lambda_call_point_deps_done_once:
                         trace_stmt.lambda_call_point_deps_done_once = True
@@ -430,14 +448,39 @@ class TraceManager(SliceTraceManager):
 
     def resolve_store_data_for_target(
         self, target: Union[str, int, ast.AST], frame: FrameType
-    ) -> Tuple[Scope, AttrSubVal, Any, bool]:
+    ) -> Tuple[Scope, AttrSubVal, Any, bool, Set[DataSymbol]]:
         target = self._partial_resolve_ref(target)
         if isinstance(target, str):
-            obj = frame.f_locals[target]
-            return self.cur_frame_original_scope, target, obj, False
+            scope = self.cur_frame_original_scope
+            lut = frame.f_locals
+            if scope.symtab is not None:
+                try:
+                    target_sym = scope.symtab.lookup(target)
+                    # this nonsense is necessary because the "is_nonlocal" method
+                    # is not available on Python <= 3.7;
+                    # the below check seems to work consistently across all Python versions
+                    is_nonlocal = getattr(
+                        target_sym, 'is_nonlocal',
+                        lambda: not target_sym.is_global()
+                                    and target_sym.is_assigned()
+                                    and target_sym.is_free()
+                    )()
+                    if is_nonlocal:
+                        scope = scope.parent_scope
+                    elif target_sym.is_global():
+                        lut = frame.f_globals
+                        scope = nbs().global_scope
+                except KeyError:
+                    pass
+            try:
+                obj = lut[target]
+            except KeyError:
+                obj = frame.f_globals[target]
+                scope = nbs().global_scope
+            return scope, target, obj, False, set()
         (
             scope, obj, attr_or_sub, is_subscript
-        ) = self.node_id_to_saved_store_data[target]
+        ) = self.node_id_to_saved_store_data.pop(target)
         attr_or_sub_obj = nbs().retrieve_namespace_attr_or_sub(obj, attr_or_sub, is_subscript)
         if attr_or_sub_obj is None:
             scope_to_use = scope
@@ -446,7 +489,13 @@ class TraceManager(SliceTraceManager):
         if scope_to_use is None:
             # Nobody before `scope` has it, so we'll insert it at this level
             scope_to_use = scope
-        return scope_to_use, attr_or_sub, attr_or_sub_obj, is_subscript
+        return (
+            scope_to_use,
+            attr_or_sub,
+            attr_or_sub_obj,
+            is_subscript,
+            self.node_id_to_saved_live_subscript_refs.pop(target, set()),
+        )
 
     def resolve_del_data_for_target(
         self, target: Union[str, int, ast.AST]
@@ -480,7 +529,7 @@ class TraceManager(SliceTraceManager):
             data_symbols.update(self.resolve_loaded_symbols(ref))
         return data_symbols
 
-    def _get_namespace_for_obj(self, obj: Any, obj_name: Optional[str] = None) -> NamespaceScope:
+    def _get_namespace_for_obj(self, obj: Any, obj_name: Optional[str] = None) -> Namespace:
         obj_id = id(obj)
         ns = nbs().namespaces.get(obj_id, None)
         if ns is not None:
@@ -500,7 +549,7 @@ class TraceManager(SliceTraceManager):
                 scope_name = nbs().get_first_full_symbol(obj_id).name if obj_name is None else obj_name
             except AttributeError:
                 scope_name = '<unknown namespace>'
-            ns = NamespaceScope(obj, scope_name, parent_scope=None)
+            ns = Namespace(obj, scope_name, parent_scope=None)
         # FIXME: brittle strategy for determining parent scope of obj
         if ns.parent_scope is None:
             if (
@@ -564,6 +613,14 @@ class TraceManager(SliceTraceManager):
     @register_handler(TraceEvent.init_cell)
     def init_cell(self, _obj, _node_id, frame: FrameType, _event, cell_id: Union[str, int], **__):
         nbs().set_name_to_cell_num_mapping(frame)
+        # needs to happen after stmt inserting has already happened
+        for flag_name in nbs().loop_iter_flag_names:
+            setattr(builtins, flag_name, False)
+
+    @register_handler(TraceEvent.after_loop_iter)
+    def after_loop_iter(self, _obj, loop_node_id: NodeId, *_, **__):
+        looped_once_flag_name = make_loop_iter_flag_name(loop_node_id)
+        setattr(builtins, looped_once_flag_name, True)
 
     @register_handler(TraceEvent.after_assign_rhs)
     @skip_when_tracing_disabled
@@ -572,7 +629,11 @@ class TraceManager(SliceTraceManager):
 
     @register_handler(TraceEvent.subscript_slice)
     @skip_when_tracing_disabled
-    def subscript_slice(self, _obj: Any, slice_node_id: NodeId, *__, **___):
+    def subscript_slice(self, _obj: Any, node_id: NodeId, *__, **___):
+        node = nbs().ast_node_by_id.get(node_id, None)
+        if node is None:
+            return
+        slice_node_id = id(cast(ast.Subscript, node).slice)
         live, _ = compute_live_dead_symbol_refs(
             nbs().ast_node_by_id[slice_node_id], scope=self.cur_frame_original_scope
         )
@@ -587,6 +648,7 @@ class TraceManager(SliceTraceManager):
             # instead just check for length-1 "chains" that are just fn calls
             if isinstance(first_in_chain, CallPoint) and len(ref.symbols) == 1:
                 subscript_live_refs.append(first_in_chain.symbol)
+        self.node_id_to_saved_live_subscript_refs[node_id] = self.resolve_symbols(set(subscript_live_refs))
         Timestamp.update_usage_info(
             self.cur_frame_original_scope.lookup_data_symbol_by_name(ref) for ref in subscript_live_refs
         )
@@ -619,10 +681,19 @@ class TraceManager(SliceTraceManager):
         if sym_for_obj is None and obj_name is not None:
             sym_for_obj = self.active_scope.lookup_data_symbol_by_name_this_indentation(obj_name)
 
-        if sym_for_obj is not None:
-            sym_for_obj.update_usage_info(exclude_ns=True)
-
+        scope = self._get_namespace_for_obj(obj, obj_name=obj_name)
         is_subscript = (event == TraceEvent.subscript)
+        if sym_for_obj is not None:
+            try:
+                data_sym = scope.lookup_data_symbol_by_name_this_indentation(
+                    attr_or_subscript, is_subscript=is_subscript, skip_cloned_lookup=True,
+                )
+            except TypeError:
+                data_sym = None
+            if data_sym is None:
+                sym_for_obj.update_usage_info()
+            else:
+                sym_for_obj.update_usage_info(exclude_ns=True)
 
         obj_id = id(obj)
         if self.top_level_node_id_for_chain is None:
@@ -630,8 +701,6 @@ class TraceManager(SliceTraceManager):
         if self.first_obj_id_in_chain is None:
             self.first_obj_id_in_chain = obj_id
 
-        scope = self._get_namespace_for_obj(obj, obj_name=obj_name)
-
         try:
             if isinstance(attr_or_subscript, tuple):
                 if not all(isinstance(v, (str, int)) for v in attr_or_subscript):
@@ -690,21 +759,23 @@ class TraceManager(SliceTraceManager):
             obj_type = next(iter(nbs().aliases[obj_id])).obj_type
         is_excepted_mutation = False
         is_excepted_non_mutation = False
-        if retval is not None and id(retval) != obj_id:
-            # doesn't look like something we can trace, but it also
-            # doesn't look like something that mutates the caller, since
-            # the return value is not None and it's not the caller object
-            if (obj_id, method_name) in METHODS_WITH_MUTATION_EVEN_FOR_NON_NULL_RETURN:
-                is_excepted_mutation = True
-            else:
-                return
-        if not is_excepted_mutation:
-            if retval is None:
-                is_excepted_non_mutation = (obj_id, method_name) in METHODS_WITHOUT_MUTATION_EVEN_FOR_NULL_RETURN
-            if is_excepted_non_mutation or obj_type is None or id(obj_type) in nbs().aliases:
-                # the calling obj looks like something that we can trace;
-                # no need to process the call as a possible mutation
-                return
+        if isinstance(mutation_event, StandardMutation):
+            # only look for exceptions for standard mutations; other cases are handled elsewhere
+            if retval is not None and id(retval) != obj_id:
+                # doesn't look like something we can trace, but it also
+                # doesn't look like something that mutates the caller, since
+                # the return value is not None and it's not the caller object
+                if (obj_id, method_name) in METHODS_WITH_MUTATION_EVEN_FOR_NON_NULL_RETURN:
+                    is_excepted_mutation = True
+                else:
+                    return
+            if not is_excepted_mutation:
+                if retval is None:
+                    is_excepted_non_mutation = (obj_id, method_name) in METHODS_WITHOUT_MUTATION_EVEN_FOR_NULL_RETURN
+                if is_excepted_non_mutation or obj_type is None or id(obj_type) in nbs().aliases:
+                    # the calling obj looks like something that we can trace;
+                    # no need to process the call as a possible mutation
+                    return
         arg_dsyms: Set[DataSymbol] = set()
         arg_dsyms = arg_dsyms.union(*recorded_arg_dsyms)
         if isinstance(mutation_event, StandardMutation):
@@ -736,8 +807,6 @@ class TraceManager(SliceTraceManager):
                         mutation_event = ArgMutate()
             except:
                 pass
-        elif isinstance(mutation_event, ListInsert):
-            mutation_event.insert_pos = recorded_arg_objs[0]
         self.mutations.append((obj_id, mutation_event, arg_dsyms, recorded_arg_objs))
 
     @register_handler(TraceEvent.after_complex_symbol)
@@ -772,6 +841,19 @@ class TraceManager(SliceTraceManager):
         if mut_cand is None:
             return
 
+        if isinstance(mut_cand[1], (ListInsert, ListPop, ListRemove)) and mut_cand[1].pos is None and self.num_args_seen == 1:
+            try:
+                if isinstance(mut_cand[1], ListRemove):
+                    mut_obj = mut_cand[0][0]
+                    for i in range(len(mut_obj)):
+                        if mut_obj[i] == arg_obj:
+                            mut_cand[1].pos = i
+                            break
+                else:
+                    mut_cand[1].pos = arg_obj
+            except:
+                pass
+
         if isinstance(arg_node, ast.Name):
             assert self.active_scope is self.cur_frame_original_scope
             arg_dsym = self.active_scope.lookup_data_symbol_by_name(arg_node.id)
@@ -783,14 +865,9 @@ class TraceManager(SliceTraceManager):
         mut_cand[-1].append(arg_obj)
 
     def _save_mutation_candidate(self, obj: Any, method_name: Optional[str], obj_name: Optional[str] = None) -> None:
-        mutation_event: MutationEvent = StandardMutation()
-        if isinstance(obj, list):
-            if method_name == 'append':
-                mutation_event = ListAppend()
-            elif method_name == 'extend':
-                mutation_event = ListExtend(len(obj))
-            elif method_name == 'insert':
-                mutation_event = ListInsert()
+        mutation_event = resolve_mutating_method(obj, method_name)
+        if mutation_event is None or isinstance(mutation_event, MutatingMethodEventNotYetImplemented):
+            mutation_event = StandardMutation()
         self.mutation_candidate = ((obj, obj_name, method_name), mutation_event, [], [])
 
     @register_handler(TraceEvent.before_call)
@@ -803,6 +880,7 @@ class TraceManager(SliceTraceManager):
             _, obj, attr_or_subscript, is_subscript, *_, obj_name = self.saved_complex_symbol_load_data
         if obj is not None and is_subscript is not None:
             if is_subscript:
+                # TODO: need to do this also for chained calls, e.g. f()()
                 method_name = None
             else:
                 assert isinstance(attr_or_subscript, str)
@@ -823,15 +901,18 @@ class TraceManager(SliceTraceManager):
                 # if tracing gets reenabled here instead of at the 'before_stmt' handler, then we're still
                 # at the same module stmt as when tracing was disabled, and we still have a 'return' to trace
                 self.call_depth = 1
-                while len(self.call_stack) > 0:
-                    self.call_stack.pop()
+                self.call_stack.clear()
+                self.lexical_call_stack.clear()
 
         if not tracing_will_be_enabled_by_end:
             return
 
         # no need to reset active scope here;
         # that will happen in the 'after chain' handler
-        self.lexical_call_stack.pop()
+
+        if len(self.lexical_call_stack) > 0:
+            # skip / give up if tracing was recently reenabled
+            self.lexical_call_stack.pop()
         self.prev_node_id_in_cur_frame_lexical = None
         self._process_possible_mutation(retval)
 
@@ -844,7 +925,7 @@ class TraceManager(SliceTraceManager):
     def before_literal(self, *_, **__):
         parent_scope = self.active_literal_scope or self.cur_frame_original_scope
         with self.lexical_literal_stack.push():
-            self.active_literal_scope = NamespaceScope(None, NamespaceScope.ANONYMOUS, parent_scope)
+            self.active_literal_scope = Namespace(None, Namespace.ANONYMOUS, parent_scope)
 
     @register_handler((TraceEvent.after_dict_literal, TraceEvent.after_list_literal, TraceEvent.after_tuple_literal))
     @skip_when_tracing_disabled
@@ -854,6 +935,7 @@ class TraceManager(SliceTraceManager):
             logger.warning("create literal scope %s", self.active_literal_scope)
             starred_idx = -1
             starred_namespace = None
+            outer_deps = set()
             for (i, inner_obj), (inner_key_node, inner_val_node) in match_container_obj_or_namespace_with_literal_nodes(
                 literal, nbs().ast_node_by_id[node_id]  # type: ignore
             ):
@@ -872,7 +954,7 @@ class TraceManager(SliceTraceManager):
                 else:
                     inner_symbols = resolve_rval_symbols(inner_val_node)
                     if inner_key_node is not None:
-                        inner_symbols.update(resolve_rval_symbols(inner_key_node))
+                        outer_deps.update(resolve_rval_symbols(inner_key_node))
                 self.node_id_to_loaded_symbols.pop(id(inner_val_node), None)
                 inner_symbols.discard(None)
                 if isinstance(i, (int, str)):  # TODO: perform more general check for SupportedIndexType
@@ -883,6 +965,10 @@ class TraceManager(SliceTraceManager):
                         self.prev_trace_stmt_in_cur_frame.stmt_node,
                         is_subscript=True,
                         implicit=True,
+                        # this is necessary in case some literal object got reused,
+                        # since as of this comment (2021/08/14) we do not clear
+                        # GC'd symbols from the symbol graph
+                        propagate=False,
                     )
             self.node_id_to_loaded_literal_scope[node_id] = self.active_literal_scope
             parent_scope: Scope = self.active_literal_scope.parent_scope
@@ -890,7 +976,7 @@ class TraceManager(SliceTraceManager):
             literal_sym = parent_scope.upsert_data_symbol_for_name(
                 '<literal_sym_%d>' % id(literal),
                 literal,
-                set(),
+                outer_deps,
                 self.prev_trace_stmt_in_cur_frame.stmt_node,
                 is_anonymous=True,
                 implicit=True,
