@@ -3,13 +3,13 @@ import ast
 import logging
 from typing import TYPE_CHECKING
 
+import nbsafety.tracing.mutation_event as me
 from nbsafety.analysis.symbol_edges import get_symbol_edges
 from nbsafety.analysis.utils import stmt_contains_lval
 from nbsafety.data_model.data_symbol import DataSymbol
-from nbsafety.data_model.scope import NamespaceScope
+from nbsafety.data_model.namespace import Namespace
 from nbsafety.data_model.timestamp import Timestamp
 from nbsafety.singletons import nbs, tracer
-from nbsafety.tracing.mutation_event import ArgMutate, ListAppend, ListExtend, ListInsert
 from nbsafety.tracing.symbol_resolver import resolve_rval_symbols
 from nbsafety.tracing.utils import match_container_obj_or_namespace_with_literal_nodes
 
@@ -25,7 +25,7 @@ class TraceStatement:
     def __init__(self, frame: FrameType, stmt_node: ast.stmt):
         self.frame: FrameType = frame
         self.stmt_node = stmt_node
-        self.class_scope: Optional[NamespaceScope] = None
+        self.class_scope: Optional[Namespace] = None
         self.lambda_call_point_deps_done_once = False
         self.node_id_for_last_call: Optional[int] = None
 
@@ -49,7 +49,7 @@ class TraceStatement:
         if isinstance(self.stmt_node, ast.ClassDef):
             # classes need a new scope before the ClassDef has finished executing,
             # so we make it immediately
-            pending_ns = old_scope.make_child_scope(self.stmt_node.name, obj=NamespaceScope.PENDING_CLASS_PLACEHOLDER)
+            pending_ns = Namespace.make_child_namespace(old_scope, self.stmt_node.name)
             tracer().pending_class_namespaces.append(pending_ns)
             return pending_ns
 
@@ -80,17 +80,17 @@ class TraceStatement:
     ) -> None:
         # logger.error("upsert %s into %s", deps, tracer()._partial_resolve_ref(target))
         try:
-            scope, name, obj, is_subscript = tracer().resolve_store_data_for_target(target, self.frame)
-        except KeyError as e:
+            scope, name, obj, is_subscript, excluded_deps = tracer().resolve_store_data_for_target(target, self.frame)
+        except KeyError:
             # e.g., slices aren't implemented yet
             # use suppressed log level to avoid noise to user
             if nbs().is_develop:
-                logger.warning('keyerror for %s', ast.dump(target) if isinstance(target, ast.AST) else target)
+                logger.exception('keyerror for %s', ast.dump(target) if isinstance(target, ast.AST) else target)
             # if nbs().is_test:
             #     raise ke
             return
         upserted = scope.upsert_data_symbol_for_name(
-            name, obj, deps, self.stmt_node, is_subscript=is_subscript,
+            name, obj, deps - excluded_deps, self.stmt_node, is_subscript=is_subscript,
         )
         logger.info("sym %s upserted to scope %s has parents %s", upserted, scope, upserted.parents)
         if maybe_fixup_literal_namespace:
@@ -108,7 +108,7 @@ class TraceStatement:
 
     def _handle_starred_store_target(self, target: ast.Starred, inner_deps: List[Optional[DataSymbol]]):
         try:
-            scope, name, obj, is_subscript = tracer().resolve_store_data_for_target(target, self.frame)
+            scope, name, obj, is_subscript, _ = tracer().resolve_store_data_for_target(target, self.frame)
         except KeyError as e:
             # e.g., slices aren't implemented yet
             # use suppressed log level to avoid noise to user
@@ -116,7 +116,7 @@ class TraceStatement:
             return
         ns = nbs().namespaces.get(id(obj), None)
         if ns is None:
-            ns = NamespaceScope(obj, str(name), scope)
+            ns = Namespace(obj, str(name), scope)
         for i, inner_dep in enumerate(inner_deps):
             deps = set() if inner_dep is None else {inner_dep}
             ns.upsert_data_symbol_for_name(i, inner_dep.obj, deps, self.stmt_node, is_subscript=True)
@@ -129,7 +129,7 @@ class TraceStatement:
         )
 
     def _handle_store_target_tuple_unpack_from_namespace(
-        self, target: Union[ast.List, ast.Tuple], rhs_namespace: NamespaceScope
+        self, target: Union[ast.List, ast.Tuple], rhs_namespace: Namespace
     ):
         saved_starred_node: Optional[ast.Starred] = None
         saved_starred_deps = []
@@ -219,9 +219,9 @@ class TraceStatement:
                 self.class_scope.obj = class_ref
                 nbs().namespaces[id(class_ref)] = self.class_scope
             try:
-                scope, name, obj, is_subscript = tracer().resolve_store_data_for_target(target, self.frame)
+                scope, name, obj, is_subscript, excluded_deps = tracer().resolve_store_data_for_target(target, self.frame)
                 scope.upsert_data_symbol_for_name(
-                    name, obj, rval_deps, self.stmt_node,
+                    name, obj, rval_deps - excluded_deps, self.stmt_node,
                     overwrite=should_overwrite,
                     is_subscript=is_subscript,
                     is_function_def=is_function_def,
@@ -241,9 +241,10 @@ class TraceStatement:
                 if nbs().is_test:
                     raise e
 
-    def _handle_list_mutation(
+    # TODO: put this logic in each respective MutationEvent itself
+    def _handle_specific_mutation_type(
         self,
-        mutation_event: Union[ListAppend, ListExtend, ListInsert],
+        mutation_event: me.MutationEvent,
         mutated_obj_id: int,
         mutation_upsert_deps: Set[DataSymbol],
     ):
@@ -252,13 +253,13 @@ class TraceStatement:
         if mutated_sym is None:
             return
         mutated_obj = mutated_sym.obj
-        if isinstance(mutation_event, (ListAppend, ListExtend)):
+        if isinstance(mutation_event, (me.ListAppend, me.ListExtend)):
             for upsert_pos in range(
-                mutation_event.orig_len if isinstance(mutation_event, ListExtend) else len(mutated_obj) - 1,
+                mutation_event.orig_len if isinstance(mutation_event, me.ListExtend) else len(mutated_obj) - 1,
                 len(mutated_obj)
             ):
                 if namespace_scope is None:
-                    namespace_scope = NamespaceScope(
+                    namespace_scope = Namespace(
                         mutated_obj,
                         mutated_sym.name,
                         parent_scope=mutated_sym.containing_scope
@@ -275,27 +276,38 @@ class TraceStatement:
                     is_subscript=True,
                     propagate=False,
                 )
-        elif isinstance(mutation_event, ListInsert):
+        elif isinstance(mutation_event, me.ListInsert):
             assert mutated_obj is namespace_scope.obj
-            namespace_scope.shuffle_symbols_upward_from(mutation_event.insert_pos)
+            namespace_scope.shuffle_symbols_upward_from(mutation_event.pos)
             namespace_scope.upsert_data_symbol_for_name(
-                mutation_event.insert_pos,
-                mutated_obj[mutation_event.insert_pos],
+                mutation_event.pos,
+                mutated_obj[mutation_event.pos],
                 mutation_upsert_deps,
                 self.stmt_node,
                 overwrite=False,
                 is_subscript=True,
                 propagate=True,
             )
+        elif isinstance(mutation_event, (me.ListPop, me.ListRemove)) and mutation_event.pos is not None:
+            assert mutated_obj is namespace_scope.obj
+            namespace_scope.delete_data_symbol_for_name(mutation_event.pos, is_subscript=True)
+        elif isinstance(mutation_event, me.NamespaceClear):
+            for name in sorted(
+                (
+                    dsym.name for dsym in
+                    namespace_scope.all_data_symbols_this_indentation(exclude_class=True, is_subscript=True)
+                ), reverse=True
+            ):
+                print(name)
+                namespace_scope.delete_data_symbol_for_name(name, is_subscript=True)
 
     def handle_dependencies(self):
         if not nbs().dependency_tracking_enabled:
             return
         for mutated_obj_id, mutation_event, mutation_arg_dsyms, mutation_arg_objs in tracer().mutations:
-            propagate_to_namespace_descendents = not isinstance(mutation_event, (ListAppend, ListExtend, ListInsert))
             logger.info("mutation %s %s %s %s", mutated_obj_id, mutation_event, mutation_arg_dsyms, mutation_arg_objs)
             Timestamp.update_usage_info(mutation_arg_dsyms)
-            if isinstance(mutation_event, ArgMutate):
+            if isinstance(mutation_event, me.ArgMutate):
                 for mutated_sym in mutation_arg_dsyms:
                     if mutated_sym is None or mutated_sym.is_anonymous:
                         continue
@@ -307,18 +319,21 @@ class TraceStatement:
             # NOTE: this next block is necessary to ensure that we add the argument as a namespace child
             # of the mutated symbol. This helps to avoid propagating through to dependency children that are
             # themselves namespace children.
-            if isinstance(mutation_event, (ListAppend, ListExtend, ListInsert)):
+            should_propagate = True
+            if not isinstance(mutation_event, (me.MutatingMethodEventNotYetImplemented, me.StandardMutation)):
+                should_propagate = False
                 mutation_upsert_deps: Set[DataSymbol] = set()
-                if isinstance(mutation_event, (ListAppend, ListInsert)):
+                if isinstance(mutation_event, (me.ListAppend, me.ListInsert)):
                     mutation_arg_dsyms, mutation_upsert_deps = mutation_upsert_deps, mutation_arg_dsyms
-                self._handle_list_mutation(mutation_event, mutated_obj_id, mutation_upsert_deps)
+                self._handle_specific_mutation_type(mutation_event, mutated_obj_id, mutation_upsert_deps)
             Timestamp.update_usage_info(nbs().aliases[mutated_obj_id])
             for mutated_sym in nbs().aliases[mutated_obj_id]:
                 mutated_sym.update_deps(
                     mutation_arg_dsyms,
                     overwrite=False,
                     mutated=True,
-                    propagate_to_namespace_descendents=propagate_to_namespace_descendents,
+                    propagate_to_namespace_descendents=should_propagate,
+                    refresh=should_propagate,
                 )
         if self._contains_lval():
             self._make_lval_data_symbols()
