@@ -32,11 +32,9 @@ def override_settings(**kwargs):
 
 def run_cell(cell, cell_id=None, **kwargs):
     """Mocks the `change active cell` portion of the comm protocol"""
-    if cell_id is not None:
-        nbs().handle({
-            'type': 'change_active_cell',
-            'active_cell_id': cell_id
-        })
+    if cell_id is None:
+        cell_id = nbs().cell_counter()
+    nbs().set_active_cell(cell_id)
     run_cell_(cell, **kwargs)
 
 
@@ -157,7 +155,6 @@ for foo in lst:
     assert response['fresh_cells'] == [2, 3, 5] + ([6] if force_subscript_symbol_creation else [])
 
 
-@skipif_known_failing
 def test_no_freshness_for_alias_assignment_post_mutation():
     cells = {
         0: 'x = []',
@@ -238,7 +235,7 @@ def test_updated_namespace_after_subscript_dep_removed():
     assert response['fresh_cells'] == [], 'got %s' % response['fresh_cells']
 
 
-def test_equal_list_update_does_not_induce_fresh_cell():
+def test_equal_list_update_does_induce_fresh_cell():
     cells = {
         0: 'x = ["f"] + ["o"] * 10',
         1: 'y = x + list("bar")',
@@ -248,14 +245,31 @@ def test_equal_list_update_does_not_induce_fresh_cell():
     run_all_cells(cells)
     response = nbs().check_and_link_multiple_cells(cells)
     assert response['stale_cells'] == []
-    assert response['fresh_cells'] == []
+    assert response['fresh_cells'] == [2]
+    run_cell('y = ("f",)', 4)
+    response = nbs().check_and_link_multiple_cells(cells)
+    assert response['stale_cells'] == []
+    assert response['fresh_cells'] == [2, 3]
+
+
+def test_equal_list_update_does_induce_fresh_cell_LITERAL_WITH_F_IS_REUSED_ON_UBUNTU_20_04_2_PYTHON_3_8_11():
+    cells = {
+        0: 'x = ["f"] + ["o"] * 10',
+        1: 'y = x + list("bar")',
+        2: 'logging.info(y)',
+        3: 'y = list("".join(y))',
+    }
+    run_all_cells(cells)
+    response = nbs().check_and_link_multiple_cells(cells)
+    assert response['stale_cells'] == []
+    assert response['fresh_cells'] == [2]
     run_cell('y = ["f"]', 4)
     response = nbs().check_and_link_multiple_cells(cells)
     assert response['stale_cells'] == []
     assert response['fresh_cells'] == [2, 3]
 
 
-def test_equal_dict_update_does_not_induce_fresh_cell():
+def test_equal_dict_update_does_induce_fresh_cell():
     cells = {
         0: 'x = {"foo": 42, "bar": 43}',
         1: 'y = dict(set(x.items()) | set({"baz": 44}.items()))',
@@ -265,7 +279,7 @@ def test_equal_dict_update_does_not_induce_fresh_cell():
     run_all_cells(cells)
     response = nbs().check_and_link_multiple_cells(cells)
     assert response['stale_cells'] == []
-    assert response['fresh_cells'] == [], 'got %s' % response['fresh_cells']
+    assert response['fresh_cells'] == [2], 'got %s' % response['fresh_cells']
     run_cell('y = {"foo": 99}', 4)
     response = nbs().check_and_link_multiple_cells(cells)
     assert response['stale_cells'] == []
@@ -286,7 +300,7 @@ def test_list_append():
     run_cell('lst[1] += 42', 4)
     response = nbs().check_and_link_multiple_cells(cells)
     assert response['stale_cells'] == [2]
-    assert response['fresh_cells'] == [1, 3]
+    assert response['fresh_cells'] == [1]
 
 
 def test_list_extend():
@@ -303,7 +317,7 @@ def test_list_extend():
     run_cell('lst[1] += 42', 4)
     response = nbs().check_and_link_multiple_cells(cells)
     assert response['stale_cells'] == [2]
-    assert response['fresh_cells'] == [1, 3]
+    assert response['fresh_cells'] == [1]
 
 
 def test_implicit_subscript_symbol_does_not_bump_ts():
@@ -331,7 +345,7 @@ def test_liveness_skipped_for_simple_assignment_involving_aliases():
     run_cell('lst = [1, 2, 3, 4]', 3)
     response = nbs().check_and_link_multiple_cells(cells)
     assert response['stale_cells'] == []
-    assert response['fresh_cells'] == [1], 'got %s' % response['fresh_cells']
+    assert response['fresh_cells'] == [1, 2], 'got %s' % response['fresh_cells']
 
 
 def test_incorrect_object_not_used_for_argument_symbols():
@@ -383,7 +397,7 @@ def test_list_insert():
     assert response['fresh_cells'] == [4, 5, 6, 7], 'got %s' % response['fresh_cells']
 
 
-def test_list_delete():
+def _test_list_delete_helper(last_cell):
     cells = {
         0: 'lst = [0, 1, 2, 3, 3, 4, 5, 6]',
         1: 'logging.info(lst[0])',
@@ -396,7 +410,7 @@ def test_list_delete():
         8: 'logging.info(lst[7])',
         9: 'x = lst[6] + 42',
         10: 'logging.info(x)',
-        11: 'del lst[3]',
+        11: last_cell,
     }
     run_all_cells(cells)
     response = nbs().check_and_link_multiple_cells(cells)
@@ -406,3 +420,64 @@ def test_list_delete():
     # TODO: ideally cell 8 would be considered unsafe since the last entry
     #  no longer exists
     assert response['fresh_cells'] == [4, 5, 6, 7, 9], 'got %s' % response['fresh_cells']
+
+
+def test_list_delete():
+    _test_list_delete_helper('del lst[3]')
+
+
+def test_list_pop():
+    _test_list_delete_helper('lst.pop(3)')
+
+
+def test_list_remove():
+    _test_list_delete_helper('lst.remove(3)')
+
+
+def test_list_clear():
+    cells = {
+        0: 'lst = [0]',
+        1: 'logging.info(lst[0])',
+        2: 'logging.info(lst)',
+    }
+    run_all_cells(cells)
+    response = nbs().check_and_link_multiple_cells(cells)
+    assert response['stale_cells'] == []
+    assert response['fresh_cells'] == []
+    run_cell('lst.clear()', 3)
+    response = nbs().check_and_link_multiple_cells(cells)
+    assert response['stale_cells'] == []
+    assert response['fresh_cells'] == [2]
+
+
+def test_dict_clear():
+    cells = {
+        0: 'd = {0: 0}',
+        1: 'logging.info(d[0])',
+        2: 'logging.info(d)',
+    }
+    run_all_cells(cells)
+    response = nbs().check_and_link_multiple_cells(cells)
+    assert response['stale_cells'] == []
+    assert response['fresh_cells'] == []
+    run_cell('d.clear()', 3)
+    response = nbs().check_and_link_multiple_cells(cells)
+    assert response['stale_cells'] == []
+    assert response['fresh_cells'] == [2]
+
+
+def test_adhoc_pandas_series_update():
+    cells = {
+        0: 'import pandas as pd',
+        1: 'df = pd.DataFrame({1: [2, 3], 4: [5, 7]})',
+        2: 'df["foo"] = [8, 9]',
+        3: 'df.foo.dropna(inplace=True)',
+    }
+    run_all_cells(cells)
+    response = nbs().check_and_link_multiple_cells(cells)
+    assert response['stale_cells'] == []
+    assert response['fresh_cells'] == []
+    run_cell('df["foo"] = [8, 9]')
+    response = nbs().check_and_link_multiple_cells(cells)
+    assert response['stale_cells'] == []
+    assert response['fresh_cells'] == [3]
