@@ -249,18 +249,18 @@ def test_basic_assignment():
     run_cell('b = 2')
     run_cell('c = a+b')
     run_cell('d = c+1')
-    run_cell('logging.info(a,b,c,d)')
+    run_cell('logging.info("%s, %s, %s, %s", a, b, c, d)')
     # redefine a here but not c and d
     run_cell('a = 7')
-    run_cell('logging.info(a,b,c,d)')
+    run_cell('logging.info("%s, %s, %s, %s", a, b, c, d)')
     assert_detected("Did not detect that c's reference was changed")
 
     run_cell('c = a+b')
-    run_cell('logging.info(a,b,c,d)')
+    run_cell('logging.info("%s, %s, %s, %s", a, b, c, d)')
     assert_detected("Did not detect that d's reference was changed")
 
     run_cell('d = c+1')
-    run_cell('logging.info(a,b,c,d)')
+    run_cell('logging.info("%s, %s, %s, %s", a, b, c, d)')
     assert_not_detected("There should be no more dependency issue")
 
 
@@ -546,19 +546,19 @@ def func():
     run_cell('x = 7')
     run_cell('y = x')
     run_cell('z = func')
-    run_cell('logging.info(y,z())')
+    run_cell('logging.info("%s, %s", y, z())')
 
     # change x inside of the function, but not x outside of the function
     run_cell('def func():\n    x = 10')
-    run_cell('logging.info(y,z())')
+    run_cell('logging.info("%s, %s", y, z())')
     assert_detected("Did not detect the dependency change in the function")
 
     run_cell('y = x')
-    run_cell('logging.info(y,z())')
+    run_cell('logging.info("%s, %s", y, z())')
     assert_detected("Updating y should not solve the dependency change inside of function func")
 
     run_cell('z = func')
-    run_cell('logging.info(y,z())')
+    run_cell('logging.info("%s, %s", y, z())')
     assert_not_detected("Updating z should solve the problem")
 
 
@@ -567,19 +567,19 @@ def test_variable_scope_2():
     run_cell('x = 7')
     run_cell('y = x')
     run_cell('z = func')
-    run_cell('logging.info(y,z())')
+    run_cell('logging.info("%s, %s", y, z())')
 
     # change x outside of the function, but not inside of the function
     run_cell('x = 10')
-    run_cell('logging.info(y,z())')
+    run_cell('logging.info("%s, %s", y, z())')
     assert_detected("Did not detect the dependency change outside of the function")
 
     run_cell('z = func')
-    run_cell('logging.info(y,z())')
+    run_cell('logging.info("%s, %s", y, z())')
     assert_detected("Updating z should not solve the dependency change outside of function")
 
     run_cell('y = x')
-    run_cell('logging.info(y,z())')
+    run_cell('logging.info("%s, %s", y, z())')
     assert_not_detected("Updating y should solve the problem")
 
 
@@ -1969,7 +1969,6 @@ def test_dict():
     assert_detected('`x` has dependency on old value of `d[0]`')
 
 
-@skipif_known_failing
 def test_dict_2():
     run_cell('d = {}; d[0] = 0')
     run_cell('x = d[0] + 1')
@@ -1984,7 +1983,7 @@ def test_default_dict():
     run_cell('x = d[0][0] + 1')
     run_cell('d = defaultdict(dict); d[0][0] = 42')
     run_cell('logging.info(x)')
-    assert_false_negative('`x` has dependency on old value of `d[0][0]`, but this is hard to detect b/c d was cleared')
+    assert_detected('`x` has dependency on old value of `d[0][0]`')
 
 
 def test_mutate_arg():
@@ -2211,6 +2210,42 @@ def test_list_sum_simple():
     assert_detected()
 
 
+def test_global_var():
+    run_cell('x = 0')
+    run_cell('y = x + 1')
+    run_cell('def f(): global x; x = 42')
+    run_cell('logging.info(y)')
+    assert_not_detected()
+    run_cell('f()')
+    run_cell('logging.info(y)')
+    assert_detected()
+
+
+def test_nonlocal_var():
+    cell_template = """
+z = 0
+def f():
+    x = 0
+    def g():
+        {stmt}
+        x = 42
+    global z
+    z = x
+    return g
+g = f()
+"""
+    run_cell(cell_template.format(stmt='pass'))
+    run_cell('y = z + 1')
+    run_cell('g()')
+    run_cell('logging.info(y)')
+    assert_not_detected()
+    run_cell(cell_template.format(stmt='nonlocal x'))
+    run_cell('y = z + 1')
+    run_cell('g()')
+    run_cell('logging.info(y)')
+    assert_detected()
+
+
 # TODO: where was I going with this?
 # def test_getitem_call():
 #     run_cell("""
